#include "AQM1248A.h"
#include <stdio.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include <unistd.h>
#include "esp_log.h"
#include <sys/param.h>
#include "driver/spi_master.h"
#include "driver/gpio.h"
#include <string.h>

char v_buf[128][6];
spi_device_handle_t spi;

void LCD_cmd(uint8_t cmd)
{
    esp_err_t ret;
    spi_transaction_t t;
    memset(&t, 0, sizeof(t));       			//Zero out the transaction
    t.length=8;                     			//Command is 8 bits
    t.tx_buffer=&cmd;               			//The data is the cmd itself
    ret=spi_device_polling_transmit(spi, &t);  	//Transmit!
    assert(ret==ESP_OK);            			//Should have had no issues.
}

void Init_LCD()
{
    esp_err_t ret;
    
    spi_bus_config_t buscfg={
        .miso_io_num=PIN_NUM_MISO,
        .mosi_io_num=PIN_NUM_MOSI,
        .sclk_io_num=PIN_NUM_CLK,
        .max_transfer_sz=3
    };
    
    spi_device_interface_config_t devcfg={
        .clock_speed_hz=SPI_MASTER_FREQ_20M,
        .mode=0,                        		//SPI mode 0
        .queue_size=1,                          //We want to be able to queue 7 transactions at a time
    };
    
    //Initialize the SPI bus
    ret=spi_bus_initialize(LCD_HOST, &buscfg, SPI_DMA_CH_AUTO);
    ESP_ERROR_CHECK(ret);
    
    //Attach the LCD to the SPI bus
    ret=spi_bus_add_device(LCD_HOST, &devcfg, &spi);
    ESP_ERROR_CHECK(ret);

    gpio_config_t io_conf = {};
    
    io_conf.intr_type = GPIO_INTR_DISABLE;		    				//disable interrupt
    io_conf.mode = GPIO_MODE_OUTPUT;    							//set as output mode
    io_conf.pin_bit_mask |= ((1ULL << LCD_CS) | (1ULL << LCD_RS)) ;	//bit mask of LCD_CS & LCD_RS
    io_conf.pull_down_en = 0;										//disable pull-down mode
    io_conf.pull_up_en = 0;						    				//disable pull-up mode
    gpio_config(&io_conf);    										//configure GPIO with the given settings

	gpio_set_level(LCD_CS, 1);
	gpio_set_level(LCD_RS, 1);
    vTaskDelay(50 / portTICK_PERIOD_MS);

	gpio_set_level(LCD_CS, 0);
	gpio_set_level(LCD_RS, 0);
    
    LCD_cmd(0xAE);
    LCD_cmd(0xA0);
    LCD_cmd(0xC8);
    LCD_cmd(0xA3);
    
    LCD_cmd(0x2C);
    vTaskDelay(50 / portTICK_PERIOD_MS);
    LCD_cmd(0x2E);
    vTaskDelay(50 / portTICK_PERIOD_MS);
   
    LCD_cmd(0x2F);
    
    LCD_cmd(0x23);
    LCD_cmd(0x81);
    LCD_cmd(0x1C);

    LCD_cmd(0xA4);
    LCD_cmd(0x40);
    LCD_cmd(0xA6);
    LCD_cmd(0xAF);
     
    gpio_set_level(LCD_CS,1);
}

void LCD_CLS(char data)
{
    int a,b;
    
    gpio_set_level(LCD_CS,0);
    for(b=0; b<6; b ++)
    {
        gpio_set_level(LCD_RS,0);
        LCD_cmd(0xB0+b);
        LCD_cmd(0x10);
        LCD_cmd(0x00);
    
        gpio_set_level(LCD_RS,1);   
        for(a=0; a<128; a++)
        {
            LCD_cmd(data);
            v_buf[a][b]=data;
        }
    }
    gpio_set_level(LCD_CS,1);
}

//----------------------------------------------------
//  点の描画　
//  int x_data  X positon   0 -> 128    
//  int x_data  Y positon   0 -> 48 
//  int cl      color 0: white  1:black 
//----------------------------------------------------
void LCD_PSET(int x_data, int y_data, int cl)
{

    int a,b;
    char c;
 
//  y_data
    a=y_data >> 3; b= y_data & 0x07;
    c=0x1; 
    while(b)
    {
        c <<= 1; b --; 
    }
    
    if(cl) v_buf[x_data][a] |= c;
    else 
    {
        c = ~c; v_buf[x_data][a] &= c;
    }

    gpio_set_level(LCD_CS,0);
    gpio_set_level(LCD_RS,0);    
    LCD_cmd(0xB0+a);
    c=x_data >> 4; c |= 0x10;
    LCD_cmd(c);
    c=x_data & 0xf;
    LCD_cmd(c);
    gpio_set_level(LCD_RS,1);
    LCD_cmd(v_buf[x_data][a]);
    gpio_set_level(LCD_CS,1);
}

//----------------------------------------------------
//  Fontの描画　
//  int x_data  X positon   0 -> 128    
//  int y_data  Y positon   0 -> 48
//  char c_data Data
//  int cl      color 0: white  1:black
//----------------------------------------------------
void LCD_Print_C(int x_data, int y_data, char c_data, int cl, int m_fl)
{
    int a,b,c,d,e,f;
    char s;

    a = c_data - 0x20; 
    for(b= 0; b < 5; b ++)
    {
        s = 0x1;
        for(c = 0; c < 8; c ++)
        {
            d = 0;
            if(Font[a][b] & s) d = 1;
            if(cl == 0)
            {
                if(d) d = 0;
                else d = 1;
            }
            for(e = 0; e < m_fl; e ++)
                for(f = 0; f < m_fl; f ++)
                    LCD_PSET(x_data + f,y_data + c * m_fl + e,d);
            s <<= 1; 
        }
        x_data += m_fl;
    }
    for(c = 0; c < 8*m_fl; c ++)
    {
        d = 0;
        if(cl == 0) d=1; 
        LCD_PSET(x_data,y_data + c,d);        
    }
}

//----------------------------------------------------
//  Strtの描画　
//  int x_data  X positon   0 -> 128    
//  int y_data  Y positon   0 -> 48
//  char *c_data Data
//  int cl      color 0: white  1:black
//----------------------------------------------------
void LCD_Print_Str(int x_data, int y_data, char *c_data, int cl, int m_fl)
{
    int a;
    a = strlen(c_data);
    while(a)
    {
        if(*c_data == 0xef)
        {
            c_data += 2;
            a -= 2;
        }
        LCD_Print_C(x_data,y_data,*c_data,cl,m_fl);
//        Serial.println(*c_data,HEX);
        a --; x_data += (6 *m_fl); c_data ++;
    }
}

//----------------------------------------------------
//   直線描画関数
//
//  int x0      start x
//  int y0      start y
//  int x1      end x
//  int y1      end y
//  int cl      color 0: white  1:black
//----------------------------------------------------
#define abs(a)  (((a)>0) ? (a) : -(a))
void LCD_LINE(int x0, int y0, int x1, int y1, int cl)
{
    int steep, t;
    int deltax, deltay, error;
    int x, y;
    int ystep;
    
    /// 差分の大きいほうを求める
    steep = (abs(y1 - y0) > abs(x1 - x0));
    /// ｘ、ｙの入れ替え
    if(steep)
    {
        t = x0; x0 = y0; y0 = t;
        t = x1; x1 = y1; y1 = t;
    }
    if(x0 > x1) 
    {
        t = x0; x0 = x1; x1 = t;
        t = y0; y0 = y1; y1 = t;
    }
    deltax = x1 - x0;                       // 傾き計算 
    deltay = abs(y1 - y0);
    error = 0;
    y = y0;
    /// 傾きでステップの正負を切り替え
    if(y0 < y1) ystep = 1; else ystep = -1;
    /// 直線を点で描画
    for(x=x0; x<x1+1; x++) 
    {
        if(steep) LCD_PSET(y,x,cl); else LCD_PSET(x,y,cl);
        error += deltay;
        if((error << 1) >= deltax) 
        {
            y += ystep;
            error -= deltax;
        }
    }
}
